/*	nds.c - M.A.R-T>Y. Notification Delivery Service daemon written in C
	Copyright (C) 2023-2024 hoff.industries

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <https://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>

#include <poll.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>

#include <limits.h>

#ifndef LIBMMP_AUTHOR
	#include "libmmp.h"
#endif

#ifndef LIBMARTY_AUTHOR
	#include "libmarty.h"
#endif

const int TARGET_PROTOCOL[3] = {2, 0, 0};
const int INSTANCE_VERSION[3] = {3, 1, 5};

const char INSTANCE_NAME[] = "m-cnds";
const char INSTANCE_TITLE[] = "Notification Delivery Service";

// nds is coupled to radio
const char NOTIFICATION_CLIENT[] = "radio";
const int CLIENT_LEN = sizeof(NOTIFICATION_CLIENT) / sizeof(NOTIFICATION_CLIENT[0]);

/* interval in seconds to keep
 * the last known client in memory
 * before requerying for a new one */
const int KEEP_CLIENT = 5;
const int MAX_PRIORITY = 4;

const char *HOST_PRIORITY[] = {
	"fuchsia",
	"mossdeep",
	"sunyshore"
};

const int MAX_HOST = sizeof(HOST_PRIORITY) / sizeof(char *);

struct MartyMessageHandle *create_handle_list(struct MMPacket *packet) {

	struct MartyMessageHandle *headHandle = (struct MartyMessageHandle *)malloc(sizeof(struct MartyMessageHandle));
	struct MartyMessageHandle *handle = headHandle;
	struct MartyMessageHandle *prevHandle = handle;

	int hostLen = -1;
	int formatBytes = -1;

	if (headHandle == NULL) {
		return NULL;
	}

	for (int i = 0; i < MAX_HOST; i++) {

		for (int j = 0; j < MAX_PRIORITY; j++) {

			if (i != 0 || (i == 0 && j != 0)) {
				handle = (struct MartyMessageHandle *)malloc(sizeof(struct MartyMessageHandle));
				if (handle == NULL) {

					marty_message_handle_free(headHandle, true);
					return NULL;
				}
			}

			marty_message_handle_factory(handle);

			handle->execute = mmpacket_2_heap(packet, true, 0);
			if (handle->execute == NULL) {

				marty_message_handle_free(headHandle, true);
				return NULL;
			}

			//bzero(handle->execute->bytes, sizeof(uint8_t)*MAX_SIZE);

			free(handle->execute->payloads[0]);
			handle->execute->payloads[0] = (uint8_t *)malloc(sizeof(uint8_t)*MAX_MESSAGE_ID_LEN);
			if (handle->execute->payloads[0] == NULL) {

				marty_message_handle_free(headHandle, true);
				return NULL;
			}

			bzero(handle->execute->payloads[0], sizeof(uint8_t)*MAX_MESSAGE_ID_LEN);

			free(handle->execute->payloads[1]);
			handle->execute->payloads[1] = (uint8_t *)malloc(sizeof(uint8_t)*MAX_CLIENT_ID_LEN);
			if (handle->execute->payloads[1] == NULL) {

				marty_message_handle_free(headHandle, true);
				return NULL;
			}

			bzero(handle->execute->payloads[1], sizeof(uint8_t)*MAX_CLIENT_ID_LEN);

			formatBytes = snprintf((char *)handle->execute->payloads[1], MAX_CLIENT_ID_LEN, "%d", j);
			handle->execute->payloads[1][formatBytes] = TERM_PAYLOAD;

			hostLen = strlen(HOST_PRIORITY[i]) + 1;

			free(handle->execute->payloads[2]);
			handle->execute->payloads[2] = (uint8_t *)malloc(sizeof(uint8_t)*hostLen);
			if (handle->execute->payloads[2] == NULL) {

				marty_message_handle_free(headHandle, true);
				return NULL;
			}

			bzero(handle->execute->payloads[2], sizeof(uint8_t)*hostLen);
			str_2_mmpayload((char *)HOST_PRIORITY[i], handle->execute->payloads[2]);

			free(handle->execute->payloads[3]);
			handle->execute->payloads[3] = (uint8_t *)malloc(sizeof(uint8_t)*CLIENT_LEN);
			if (handle->execute->payloads[3] == NULL) {

				marty_message_handle_free(headHandle, true);
				return NULL;
			}

			bzero(handle->execute->payloads[3], sizeof(uint8_t)*CLIENT_LEN);
			str_2_mmpayload((char *)NOTIFICATION_CLIENT, handle->execute->payloads[3]);

			prevHandle->next = handle;
			prevHandle = handle;
		}
	}

	return headHandle;
}

int main(int argc, char **argv) {
	pthread_t networkThread;
	struct MartyClientData data;
	struct MMStatus status;

	int bytesRead = 0;
	struct MMPacket txPacket;
	struct MMPacket *rxPacket = NULL;

	mmp_arg applicationStatus = MMP_APPLICATION_SUCCESS;

	char hostname[_POSIX_HOST_NAME_MAX];

	struct MartyMessageHandle *deliveryHandle = NULL;

	printf("%s[%s]%s Starting hoff.industries %s - %s v%d.%d.%d\n",
		   LoggerColours[LOG_INFO], MartyFlavour, ClearColour, INSTANCE_TITLE,
		INSTANCE_NAME, INSTANCE_VERSION[0], INSTANCE_VERSION[1], INSTANCE_VERSION[2]);

	if (marty_client_data_factory(&data) == -1) {
		marty_part_log("Client data init failure!\n", LOG_CRITICAL);
		exit(EXIT_ERR_CDATA);
	}

	mmstatus_factory(&status);
	mmpacket_factory(&txPacket, true);

	data.versionString = alloc_marty_version_string(INSTANCE_VERSION, INSTANCE_NAME);
	if (data.versionString == NULL) {
		marty_part_log("Version string allocation failure!\n", LOG_CRITICAL);
		exit(EXIT_ERR_VSTRING);
	}

	if (gethostname(hostname, _POSIX_HOST_NAME_MAX) == -1) {
		marty_part_log("Hostname get failure!\n", LOG_CRITICAL);
		exit(EXIT_ERR_HOSTNAME);
	}

	data.connType = CONN_UNIX;
	data.initParameters.auth = "";
	data.initParameters.priority = 0;
	data.initParameters.host = hostname;
	data.initParameters.name = "nds";

	if (spawn_marty_socket(&data) != MMP_ERR_NONE) {
		marty_full_log("Socket spawn failure!\n", LOG_CRITICAL, &data);
		exit(EXIT_ERR_SPAWN);
	}

	if (marty_init(&data) != MMP_ERR_NONE) {
		marty_full_log("MMP initialisation failure!\n", LOG_CRITICAL, &data);
		exit(EXIT_ERR_INIT);
	};

	marty_full_log("MMP initialisation success!\n", LOG_STARTUP, &data);

	if (pthread_create(&networkThread, NULL, marty_network_loop, (void *)&data) != 0) {

		marty_full_log("Error spawning network thread!\n", LOG_CRITICAL, &data);
		exit(EXIT_ERR_SPAWN);
	}
	marty_full_log("Network loop started.\n", LOG_STARTUP, &data);

	while (true) {

		mmpacket_reset(&txPacket, true);
		applicationStatus = MMP_APPLICATION_SUCCESS;

		if (rxPacket != NULL) {
			mmpacket_reset(rxPacket, true);
			free(rxPacket);
		}

		if (deliveryHandle != NULL) {
			marty_message_handle_free(deliveryHandle, true);
		}

		rxPacket = NULL;
		deliveryHandle = NULL;

		if (poll(data.pfds, data.nfds, -1) == -1) {
			marty_full_log("Error polling inter-thread pipes!\n", LOG_CRITICAL, &data);
			exit(EXIT_ERR_POLL);
		}

		if (data.pfds[ITC_DELIVERY_PIPE].revents != 0) {

			bytesRead = read(data.netDeliveryPipe[0], &deliveryHandle,
				sizeof(struct MartyMessageHandle *));

			if (bytesRead == 0 || bytesRead == -1) {
				marty_full_log("Error reading inter-thread pipes!\n", LOG_CRITICAL, &data);
				exit(EXIT_ERR_ITC);
			}

			if (deliveryHandle == NULL) {
				// if ITC pipe sent -2 means reconnect event occurred.
				continue;
			}

			switch (deliveryHandle->result) {
				case MMP_APPLICATION_DELIVERY:
					break;

				case MMP_APPLICATION_SUCCESS:
					marty_full_log("Another notification successfully delivered by the M.A.R-T>Y. Notification Delivery Service.\n",
								   LOG_SUCCESS, &data);

					continue;

				default:
					marty_full_log("Failed to deliver notification: application level error 0x",
								   LOG_WARNING, &data);
					printf("%x\n", deliveryHandle->result);

					continue;
			}

			if (deliveryHandle->execute == NULL) {

				marty_full_log("Failed to deliver notification: no clients up!\n", LOG_ERROR, &data);
				continue;
			}

			marty_message_handler(deliveryHandle->next, &status, &data);
			if (status.error != MMP_ERR_NONE) {

				marty_full_log("Failed to execute handle!\n", LOG_ERROR, &data);
			}

			marty_message_handle_free(deliveryHandle, false);
			deliveryHandle = NULL;
			continue;

		}

		if (data.pfds[ITC_MESSAGE_PIPE].revents == 0) {
			// since MESSAGE events are last on our list to check
			// we can continue if we don't need to handle
			continue;
		}

		bytesRead = read(data.netMessagePipe[0], &rxPacket,
			sizeof(struct MMPacket *));

		if (bytesRead == 0 || bytesRead == -1) {
			marty_full_log("Error reading inter-thread pipes!\n", LOG_CRITICAL, &data);
			exit(EXIT_ERR_ITC);
		}

		if (rxPacket == NULL) {
			marty_full_log("Received NULL pointer from ITC pipe!\n", LOG_WARNING, &data);
			continue;
		}

		deliveryHandle = create_handle_list(rxPacket);
		if (deliveryHandle == NULL) {
			marty_full_log("Error constructing handle delivery linked list!\n", LOG_ERROR, &data);
			applicationStatus = MMP_APPLICATION_EXECUTE;

			goto JMP_TX_STATUS;
		}

		marty_message_handler(deliveryHandle, &status, &data);
		if (status.error != MMP_ERR_NONE) {
			marty_full_log("Error executing delivery linked list!\n", LOG_ERROR, &data);
			applicationStatus = MMP_APPLICATION_EXECUTE;

			goto JMP_TX_STATUS;
		}
		marty_full_log("Handle executed.\n", LOG_INFO, &data);

		deliveryHandle = NULL;

		JMP_TX_STATUS:

		marty_set_application(&txPacket, &status, rxPacket->payloads[0], applicationStatus);
		if (status.error != MMP_ERR_NONE) {
			marty_full_log("Couldn't set autopacket!\n", LOG_ERROR, &data);
			continue;
		}

		mmp_serialize(&txPacket, &status);
		if (status.error != MMP_ERR_NONE) {
			marty_full_log("Couldn't serialize autopacket!\n", LOG_ERROR, &data);
			continue;
		}

		marty_net_write(&txPacket, &status, &data);
	}
}















