/*	libmarty.c - M.A.R-T>Y. C lib helper function definition file
	Copyright (C) 2023-2024 hoff.industries

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <https://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>

#include <poll.h>
#include <unistd.h>
#include <pthread.h>

#include <time.h>
#include <string.h>

#include <arpa/inet.h>
#include <netinet/in.h>

#include <sys/un.h>
#include <sys/socket.h>

#ifndef LIBMMP_AUTHOR
	#include "libmmp.h"
#endif

#include "libmarty.h"

const int LIBMARTY_VERSION[3] = {3, 1, 3};

char defaultSocket[] = "/home/mdhoff/.config/marty/entrypoint";
char testSocket[] = "/home/mdhoff/.config/marty/testpoint";
char defaultHost[] = "127.0.0.1";
char preinit[] = "__preinit__";

const char MartyColour[] = "\033[38;5;45m";
const char ClearColour[] = "\033[0m";

const char MartyFlavour[] = "M.A.R-T>Y.";
const char BracketFlavour[] = "[M.A.R-T>Y.]";

const int FlavourLen = 10;
const int BracketFlavourLen = 12;

const char *clientColours[] = {
	"\033[38;5;48m", // spring green
	"\033[38;5;220m", // gold
	"\033[38;5;27m", // dark blue
	"\033[38;5;196m", // red
	"\033[38;5;201m", // magenta
	"\033[38;5;208m", // dark orange
	"\033[38;5;124m", // dark red
	"\033[38;5;93m", // purple
	"\033[38;5;34m" // green

};

const char *LoggerColours[] = {
	"\033[38;5;48m",
	MartyColour,
	"\033[0;38;5;202m",
	"\033[38;5;124m",
	"\033[38;5;196m",
	"\033[38;5;93m",
	"",
	"\033[38;5;48m"
};

void marty_client_parameters_factory(struct MartyClientParameters *parameters) {
	parameters->auth = preinit;
	parameters->priority = 0;
	parameters->host = preinit;
	parameters->name = preinit;
}

int marty_client_data_factory(struct MartyClientData *clientData) {
	clientData->versionString = NULL;

	clientData->targetHost = defaultHost;
	clientData->targetPort = DEFAULT_PORT;
	clientData->targetSocket = defaultSocket;

	clientData->connType = CONN_TCP;
	clientData->logLevel = LOG_NOMINAL;

	if (pipe(clientData->netMessagePipe) == -1 ||
			pipe(clientData->netDeliveryPipe) == -1) {

		return -1;
	}

	clientData->nfds = 2;

	clientData->pfds[ITC_DELIVERY_PIPE].events = POLLIN;
	clientData->pfds[ITC_DELIVERY_PIPE].fd = clientData->netDeliveryPipe[0];

	clientData->pfds[ITC_MESSAGE_PIPE].events = POLLIN;
	clientData->pfds[ITC_MESSAGE_PIPE].fd = clientData->netMessagePipe[0];

	clientData->socket = -1;
	if (pthread_mutex_init(&clientData->networkLock, NULL) != 0) {
		return -1;
	}

	marty_client_parameters_factory(&clientData->initParameters);
	marty_client_parameters_factory(&clientData->myParameters);

	for (int i = 0; i < MAX_MESSAGE_ID; i++) {
		clientData->deliveryTracker[i] = NULL;
	}

	if (pthread_mutex_init(&clientData->deliveryLock, NULL) != 0) {
		return -1;
	}

	return 0;
}

void marty_client_endpoint_factory(struct MartyClientEndpoint *endpoint) {
	endpoint->endpoint = -1;
	endpoint->name = NULL;
	endpoint->min = -1;
	endpoint->max = -1;
}

void marty_message_handle_factory(struct MartyMessageHandle *handle) {
	handle->result = MMP_ARG_UNSET;
	handle->execute = NULL;
	handle->next = NULL;
}

void marty_part_log(char *message, int level) {
	printf("%s[%s]%s %s", LoggerColours[level], MartyFlavour, ClearColour, message);
}

void marty_full_log(char *message, int level, struct MartyClientData *data) {
	time_t timestamp = time(NULL);
	struct tm *timeInfo = localtime(&timestamp);

	printf("%s[%s] %s<%s> %s(%d | %s | %s)%s %d-%d-%dT%d:%d:%d %s", LoggerColours[level],
		   MartyFlavour, clientColours[C_DAEMON_COLOUR], data->versionString, MartyColour, data->myParameters.priority,
		data->myParameters.host, data->myParameters.name, ClearColour, timeInfo->tm_year + 1900,
		timeInfo->tm_mon + 1, timeInfo->tm_mday, timeInfo->tm_hour, timeInfo->tm_min, timeInfo->tm_sec, message);
}

char *alloc_marty_version_string(const int instanceVersion[3], const char *instanceName) {
	char *versionString = NULL;
	int stringSize;

	if (instanceVersion[0] > 9 || instanceVersion[1] > 9 || instanceVersion[2] > 9) {
		return NULL;
	}

	stringSize = strlen(instanceName) + 8;
	versionString = (char *)malloc(sizeof(char)*(stringSize));

	if (versionString == NULL) {
		return NULL;
	}

	if (snprintf(versionString, stringSize, "%s v%d.%d.%d", instanceName,
			instanceVersion[0], instanceVersion[1], instanceVersion[2]) < 0) {

		return NULL;
	}

	return versionString;
}

void marty_set_application(struct MMPacket *packet, struct MMStatus *status,
		uint8_t *messageId, mmp_arg argument) {

	int payloadLen = mmpayloadlen(messageId);

	mmstatus_factory(status);

	if (payloadLen > MAX_MESSAGE_ID - 1) {
		status->error = MMP_ERR_MESSAGE_ID_UNAVAILABLE;
		return;
	}

	packet->type = MMP_TYPE_APPLICATION;
	packet->argument = argument;

	packet->payloads = (uint8_t **)malloc(sizeof(uint8_t *)*MMP_PLEN_APPLICATION);
	if (packet->payloads == NULL) {
		status->error = MMP_ERR_MEMORY;
		return;
	}

	packet->payloadLen = MMP_PLEN_APPLICATION;
	packet->payloads[0] = (uint8_t *)malloc(sizeof(uint8_t)*MAX_MESSAGE_ID_LEN);
	if (packet->payloads[0] == NULL) {
		status->error = MMP_ERR_MEMORY;
		return;
	}

	for (int i = 0; i < payloadLen; i++) {
		packet->payloads[0][i] = messageId[i];
	}

	packet->payloads[0][payloadLen] = TERM_PAYLOAD;
}

void marty_net_write(struct MMPacket *txPacket, struct MMStatus *status,
		struct MartyClientData *clientData) {

	pthread_mutex_lock(&clientData->networkLock);
	write(clientData->socket, txPacket->bytes, sizeof(uint8_t)*(status->size));
	pthread_mutex_unlock(&clientData->networkLock);
}

int get_new_message_id(struct MartyClientData *data) {

	int i;

	for (i = 0; i < MAX_MESSAGE_ID; i++) {
		if (data->deliveryTracker[i] == NULL) {
			return i;
		}
	}

	return -1;
}

void marty_message_handle_free(struct MartyMessageHandle *handle, bool clearList) {

	if (handle == NULL) {
		return;
	}

	handle->result = 0;
	if (handle->execute != NULL) {

		mmpacket_reset(handle->execute, true);
		free(handle->execute);
	}

	if (clearList) {

		marty_message_handle_free(handle->next, true);
	}

	free(handle);
}

void marty_message_handler(struct MartyMessageHandle *handle,
		struct MMStatus *status, struct MartyClientData *data) {

	int messageId = -1;
	int formatBytes = -1;

	mmstatus_factory(status);
	// we use status here as the message id flag
	// so technically 0 is valid, set to -1 to be sure
	status->size = -1;

	if (handle == NULL) {
		return;
	}

	if (handle->execute == NULL) {
		status->error = MMP_ERR_UNSET_PACKET;
		return;
	}

	if (handle->execute->type != MMP_TYPE_SEND) {
		status->error = MMP_ERR_INVALID_TYPE;
		return;
	}

	pthread_mutex_lock(&data->deliveryLock);

	messageId = get_new_message_id(data);

	if (messageId == -1) {

		status->error = MMP_ERR_MESSAGE_ID_UNAVAILABLE;
		pthread_mutex_unlock(&data->deliveryLock);
		return;
	}

	if (handle->execute->payloads == NULL || handle->execute->payloadLen < MMP_PLEN_SEND) {

		status->error = MMP_ERR_INVALID_PAYLOAD_COUNT;
		pthread_mutex_unlock(&data->deliveryLock);
		return;
	}

	// WARNING: allocating enough memory for the message id (MAX_MESSAGE_ID_LEN bytes including term)
	// should be done by the caller for greater flexibility... be careful!
	formatBytes = snprintf((char *)handle->execute->payloads[0], MAX_MESSAGE_ID_LEN, "%d", messageId);
	handle->execute->payloads[0][formatBytes] = TERM_PAYLOAD;

	mmp_serialize(handle->execute, status);
	if (status->error != MMP_ERR_NONE) {

		status->size = -1;
		pthread_mutex_unlock(&data->deliveryLock);
		return;
	}

	data->deliveryTracker[messageId] = handle;
	pthread_mutex_unlock(&data->deliveryLock);

	marty_net_write(handle->execute, status, data);
	status->size = messageId;
}

enum MMPError marty_init(struct MartyClientData *data) {

	// TODO: I think there's lost memory here somewhere
	// not a big deal since this is only called on connect
	// and on reconn only if socket spawn succeeds.

	int bytesRead = 0;
	struct MMPacket txPacket;
	struct MMPacket rxPacket;
	struct MMStatus status;

	mmpacket_factory(&txPacket, true);
	mmpacket_factory(&rxPacket, true);
	mmstatus_factory(&status);

	txPacket.type = MMP_TYPE_INIT;
	txPacket.argument = MMP_ARG_INIT_INIT;
	txPacket.payloadLen = MMP_PLEN_INIT;

	txPacket.payloads = (uint8_t **)malloc(sizeof(uint8_t *)*MMP_PLEN_INIT);

	if (txPacket.payloads == NULL) {
		return MMP_ERR_MEMORY;
	}

	txPacket.payloads[0] = (uint8_t *)malloc(sizeof(uint8_t)*(strlen(
		data->initParameters.auth) + 1));
	txPacket.payloads[1] = (uint8_t *)malloc(sizeof(uint8_t)*2);
	txPacket.payloads[2] = (uint8_t *)malloc(sizeof(uint8_t)*(strlen(
		data->initParameters.host)) + 1);
	txPacket.payloads[3] = (uint8_t *)malloc(sizeof(uint8_t)*(strlen(
		data->initParameters.name) + 1));

	if (txPacket.payloads[1] == NULL || txPacket.payloads[2] == NULL ||
			txPacket.payloads[3] == NULL || txPacket.payloads[0] == NULL) {

		mmpacket_reset(&txPacket, true);
		return MMP_ERR_MEMORY;
	}

	txPacket.payloads[1][0] = data->initParameters.priority + '0';
	txPacket.payloads[1][1] = TERM_PAYLOAD;

	str_2_mmpayload(data->initParameters.auth, txPacket.payloads[0]);
	str_2_mmpayload(data->initParameters.host, txPacket.payloads[2]);
	str_2_mmpayload(data->initParameters.name, txPacket.payloads[3]);

	mmp_serialize(&txPacket, &status);
	if (status.error != MMP_ERR_NONE) {

		goto JMP_CLEANUP_RETURN;
	}

	marty_net_write(&txPacket, &status, data);

	bytesRead = read(data->socket,
		rxPacket.bytes, sizeof(uint8_t)*MAX_SIZE);

	if (bytesRead == 0 || bytesRead == -1) {

		status.error = MMP_ERR_NETWORK;
		goto JMP_CLEANUP_RETURN;
	}

	mmp_deserialize(&rxPacket, &status, 0);
	if (status.error != MMP_ERR_NONE) {

		goto JMP_CLEANUP_RETURN;
	}

	if (rxPacket.type == MMP_TYPE_INVALID &&
			rxPacket.argument == MMP_ARG_INVALID_AUTH) {

		status.error = MMP_ERR_AUTH_FAIL;
		goto JMP_CLEANUP_RETURN;
	}

	if (rxPacket.type != MMP_TYPE_INIT) {

		status.error = MMP_ERR_INVALID_TYPE;
		goto JMP_CLEANUP_RETURN;
	}

	if (rxPacket.argument != MMP_ARG_INIT_ACCEPT) {

		status.error = MMP_ERR_INIT_FAIL;
		goto JMP_CLEANUP_RETURN;
	}

	data->myParameters.auth = data->initParameters.auth;
	data->myParameters.priority = data->initParameters.priority;
	data->myParameters.host = data->initParameters.host;
	data->myParameters.name = data->initParameters.name;

	status.error = MMP_ERR_NONE;

	JMP_CLEANUP_RETURN:

	mmpacket_reset(&txPacket, true);
	mmpacket_reset(&rxPacket, true);

	return status.error;
}

enum MMPError spawn_marty_socket(struct MartyClientData *data) {

	struct sockaddr_in serverINET;
	struct sockaddr_un serverUNIX;
	bzero(&serverINET, sizeof(serverINET));
	bzero(&serverUNIX, sizeof(serverUNIX));

	if (data->connType == CONN_TCP) {

		data->socket = socket(AF_INET, SOCK_STREAM, 0);
		serverINET.sin_family = AF_INET;
		serverINET.sin_addr.s_addr = inet_addr(data->targetHost);
		serverINET.sin_port = htons(data->targetPort);

		if (connect(data->socket, (struct sockaddr *)&serverINET,
				sizeof(serverINET)) == -1) {

			return MMP_ERR_NETWORK;
		}

		return MMP_ERR_NONE;

	}

	if (data->targetSocket == NULL) {
		return MMP_ERR_INIT_FAIL;
	}

	data->socket = socket(AF_UNIX, SOCK_STREAM, 0);
	serverUNIX.sun_family = AF_UNIX;
	strcpy(serverUNIX.sun_path, data->targetSocket);

	if (connect(data->socket, (struct sockaddr *)&serverUNIX,
		sizeof(serverUNIX)) == -1) {

		return MMP_ERR_NETWORK;
	}

	return MMP_ERR_NONE;
}

void handle_reconnect(struct MartyClientData *data, int tryInterval) {

	enum MMPError error = MMP_ERR_NETWORK;

	int tries = 0;
	int tpm = 1;

	if (RECONNECT_INTERVAL < 60) {
		tpm = 60 / RECONNECT_INTERVAL;
	}

	while (error != MMP_ERR_NONE) {

		if (tries == tpm) {
			marty_full_log("Multiple attempts failed. Trying more...\n", LOG_ERROR, data);
			tries = 0;
		}

		error = spawn_marty_socket(data);

		if (error == MMP_ERR_NONE) {
			error = marty_init(data);
		}

		tries++;

		sleep(tryInterval);
	}
}


void *marty_network_loop(void *clientData) {

	int bytesRead = 0;
	int start = 0;
	int currentPacket = 0;
	int totalPackets = 0;

	long idConversion = -1;
	char *end;

	char overflowBuffer[MAX_SIZE];

	struct MartyClientData *data = (struct MartyClientData *)clientData;

	struct MMPacket rxPacket;
	struct MMPacket txPacket;

	struct MMPacket *pipePacket = NULL;

	struct MMStatus status;

	struct MartyMessageHandle *handle = NULL;

	char sendReconnect[sizeof(struct MartyMessageHandle *)] = {0};

	mmpacket_factory(&rxPacket, true);
	mmstatus_factory(&status);

	bzero(overflowBuffer, sizeof(overflowBuffer));

	while (true) {

		pipePacket = NULL;
		mmpacket_factory(&txPacket, true);

		if (start >= MAX_SIZE - 1 || rxPacket.bytes[start] == DATA_EMPTY) {

			start = 0;

			mmpacket_reset(&rxPacket, true);
			bytesRead = read(data->socket, rxPacket.bytes, MAX_SIZE);

			if (bytesRead == 0 || bytesRead == -1) {
				marty_full_log("Connection closed, reconnecting...\n", LOG_WARNING, data);

				handle_reconnect(data, RECONNECT_INTERVAL);
				marty_full_log("Reconnection success!\n", LOG_STARTUP, data);

				write(data->netDeliveryPipe[1], sendReconnect, sizeof(struct MartyMessageHandle *));
				continue;
			}
		}

		mmp_deserialize(&rxPacket, &status, start);
		if (status.error != MMP_ERR_NONE) {
			marty_full_log("Couldn't deserialize packet!\n", LOG_ERROR, data);

			for (int i = start; i < MAX_SIZE; i++) {
				start++;
				if (rxPacket.bytes[i] == TERM_PACKET) {
					break;
				}
			}

			continue;
		}

		start += status.size;

		switch (rxPacket.type) {
			case MMP_TYPE_SONAR:

				txPacket.type = MMP_TYPE_SONAR;
				break;

			case MMP_TYPE_SEND:

				marty_full_log("Direct message received.\n", LOG_INFO, data);
				pipePacket = mmpacket_2_heap(&rxPacket, false, start - status.size);
				if (pipePacket == NULL) {
					marty_full_log("Error allocating message to heap!\n", LOG_ERROR, data);
					continue;
				}

				write(data->netMessagePipe[1], &pipePacket, sizeof(struct MMPacket *));
				continue;

			case MMP_TYPE_APPLICATION:

				marty_full_log("Application message received.\n", LOG_INFO, data);
				rxPacket.payloads[0][mmpayloadlen(rxPacket.payloads[0])] = '\0';
				idConversion = strtol((char *)rxPacket.payloads[0], &end, 10);

				if (idConversion < 0 || idConversion > MAX_MESSAGE_ID) {
					marty_full_log("Error parsing application message id!\n", LOG_ERROR, data);
					continue;
				}

				pthread_mutex_lock(&data->deliveryLock);
				handle = data->deliveryTracker[idConversion];

				if (handle == NULL) {

					marty_full_log("Error mapping application message id!\n", LOG_ERROR, data);
					pthread_mutex_unlock(&data->deliveryLock);
					continue;
				}

				data->deliveryTracker[idConversion] = NULL;
				pthread_mutex_unlock(&data->deliveryLock);

				handle->result = rxPacket.argument;
				write(data->netDeliveryPipe[1],&handle, sizeof(struct MartyMessageHandle *));

				continue;

			default:
				continue;
		}

		mmp_serialize(&txPacket, &status);
		if (status.error != MMP_ERR_NONE) {
			marty_full_log("Couldn't serialize autopacket!\n", LOG_ERROR, data);
			continue;
		}

		marty_net_write(&txPacket, &status, data);
	}
}
