/*	libmmp.c - M.A.R-T>Y. C lib protocol implementation definition file
	Copyright (C) 2023-2024 hoff.industries

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <https://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>

#include <stdbool.h>

#include <string.h>
#include <strings.h>

#include "libmmp.h"

#define MAX_TYPES 8

const int MMP_LIB_VERSION[3] = {0, 0, 2};
const int MMP_IMPLEMENTS[3] = {2, 0, 0};

const int PayloadLengths[MAX_TYPES] = {
	MMP_PLEN_UNSET,
	MMP_PLEN_INIT,
	MMP_PLEN_INVALID,
	MMP_PLEN_CONTERM,
	MMP_PLEN_SONAR,
	MMP_PLEN_SEND,
	MMP_PLEN_APPLICATION,
	MMP_PLEN_QUERY
};

const int ArgumentRange[MAX_TYPES] = {
	MMP_ARG_UNSET,
	MMP_ARG_INIT_ACCEPT,
	MMP_ARG_CONTERM_AUTH,
	MMP_ARG_UNSET,
	MMP_ARG_INVALID_MESSAGE_ID,
	MMP_SEND_BROADCAST,
	MMP_APPLICATION_EXECUTE,
	MMP_QUERY_NOT_FOUND
};

bool mmp_compatible(const int programVersion[3]) {
	return (MMP_IMPLEMENTS[0] == programVersion[0] &&
		MMP_IMPLEMENTS[1] == programVersion[1] &&
			MMP_IMPLEMENTS[2] == programVersion[2]);
}


void print_mmpacket(struct MMPacket *packet) {
	for (int i = 0; i < MAX_SIZE; i++) {

		printf("%x ", packet->bytes[i]);

		if (packet->bytes[i] == TERM_PACKET) {
			break;
		}
	}

	printf("\n");
}

void mmpacket_factory(struct MMPacket *packet, bool zero) {
	packet->type = MMP_TYPE_UNSET;
	packet->argument = MMP_ARG_UNSET;
	packet->payloadLen = 0;

	packet->payloads = NULL;
	if (zero) {
		bzero(packet->bytes, MAX_SIZE);
	}
}

int mmpayloadlen(uint8_t *payload) {
	int i;

	for (i = 0; payload[i] != TERM_PAYLOAD; i++);

	return i;
}

void str_2_mmpayload(char *string, uint8_t *payload) {

	int i;

	for (i = 0; i < strlen(string); i++) {
		payload[i] = string[i];
	}

	payload[i] = TERM_PAYLOAD;
}

void mmpayload_2_str(uint8_t *payload, char *string) {

	int i;

	for (i = 0; i < mmpayloadlen(payload); i++) {
		string[i] = payload[i];
	}

	string[i] = '\0';
}

void mmpacket_reset(struct MMPacket *packet, bool zero) {

	// WARNING: this function assumes a valid packet

	if (packet->payloads == NULL) {
		mmpacket_factory(packet, zero);
		return;
	}

	for (int i = 0; i < packet->payloadLen; i++) {
		if (packet->payloads[i] != NULL) {

			free(packet->payloads[i]);
			packet->payloads[i] = NULL;
		}
	}

	free(packet->payloads);
	mmpacket_factory(packet, zero);
}

void mmstatus_factory(struct MMStatus *status) {
	status->size = 0;
	status->error = MMP_ERR_NONE;
}

struct MMPacket *mmpacket_2_heap(struct MMPacket *packet, bool doSerial, int start) {

	struct MMStatus status;
	struct MMPacket *heapPacket = (struct MMPacket *)malloc(sizeof(struct MMPacket));

	if (heapPacket == NULL) {
		return NULL;
	}

	mmpacket_factory(heapPacket, true);
	mmstatus_factory(&status);

	if (doSerial) {
		mmp_serialize(packet, &status);

		if (status.error != MMP_ERR_NONE) {

			mmpacket_reset(heapPacket, true);
			free(heapPacket);
			return NULL;
		}
	}

	for (int i = 0; i < MAX_SIZE; i++) {
		heapPacket->bytes[i] = packet->bytes[i];
	}

	mmp_deserialize(heapPacket, &status, start);
	if (status.error != MMP_ERR_NONE) {

		mmpacket_reset(heapPacket, true);
		free(heapPacket);
		return NULL;
	}

	return heapPacket;
}

void mmp_serialize(struct MMPacket *packet, struct MMStatus *status) {

	int payloadLength = -1;
	int packetIndex = PKT_IDX_PAYLOAD_START;

	mmstatus_factory(status);
	bzero(packet->bytes, sizeof(packet->bytes));

	if (packet->type == MMP_TYPE_UNSET) {
		status->error = MMP_ERR_UNSET_MESSAGE_TYPE;
		return;
	}

	if (packet->type >= MAX_TYPES) {
		status->error = MMP_ERR_INVALID_TYPE;
		return;
	}

	if (packet->argument > ArgumentRange[packet->type]) {
		status->error = MMP_ERR_INVALID_ARGUMENT;
		return;
	}

	payloadLength = PayloadLengths[packet->type];
	if (payloadLength != 0 && packet->payloads == NULL) {
		status->error = MMP_ERR_UNSET_PAYLOADS;
		return;
	}

	if (payloadLength == 0 && packet->payloads != NULL) {
		status->error = MMP_ERR_INVALID_PAYLOAD_COUNT;
		return;
	}

	if (payloadLength > packet->payloadLen) {
		status->error = MMP_ERR_INVALID_PAYLOAD_COUNT;
		return;
	}

	packet->bytes[0] = packet->type;
	packet->bytes[1] = packet->argument >> 8;
	packet->bytes[2] = packet->argument & 0xFF;

	for (int i = 0; i < packet->payloadLen; i++) {

		if (packet->payloads[i] == NULL) {

			status->error = MMP_ERR_UNSET_PAYLOADS;
			return;
		}

		for (int j = 0; j < MAX_SIZE - MIN_SIZE; j++) {
			if (packetIndex >= MAX_SIZE) {

				status->error = MMP_ERR_INVALID_PACKET_LENGTH;
				return;
			}

			packet->bytes[packetIndex] = packet->payloads[i][j];
			packetIndex++;

			if (packet->payloads[i][j] == TERM_PAYLOAD) {
				break;
			}
		}
	}

	if (packetIndex >= MAX_SIZE) {
		status->error = MMP_ERR_INVALID_PACKET_LENGTH;
		return;
	}

	packet->bytes[packetIndex] = TERM_PACKET;
	status->size = ++packetIndex;
}

void mmp_deserialize(struct MMPacket *packet, struct MMStatus *status, int start) {

	int payloadLength = -1;
	int payloadSize = 0;

	int *payloadSizes = NULL;
	int packetIndex = start + PKT_IDX_PAYLOAD_START;

	mmstatus_factory(status);
	mmpacket_reset(packet, false);

	packet->type = packet->bytes[start];
	packet->argument = packet->bytes[start + 1] << 8 | packet->bytes[start + 2];

	if (packet->type >= MAX_TYPES) {
		status->error = MMP_ERR_INVALID_TYPE;
		return;
	}

	if (packet->argument > ArgumentRange[packet->type]) {
		status->error = MMP_ERR_INVALID_ARGUMENT;
		return;
	}

	for (int i = packetIndex; i < MAX_SIZE; i++) {
		if (packet->bytes[i] == TERM_PAYLOAD) {
			packet->payloadLen++;
		}

		if (packet->bytes[i] == TERM_PACKET) {
			break;
		}
	}

	payloadLength = PayloadLengths[packet->type];

	if (packet->payloadLen < payloadLength) {
		status->error = MMP_ERR_INVALID_PAYLOAD_COUNT;
		return;
	}

	if (packet->payloadLen == 0) {
		status->size = MIN_SIZE;
		return;
	}

	payloadSizes = (int *)malloc(sizeof(int)*packet->payloadLen);
	payloadLength = 0;

	for (int i = packetIndex; i < MAX_SIZE; i++) {

		if (packet->bytes[i] == TERM_PAYLOAD) {

			payloadSizes[payloadLength] = ++payloadSize;
			payloadLength++;
			payloadSize = 0;
			continue;
		}

		if (packet->bytes[i] == TERM_PACKET) {
			if (packet->payloadLen < payloadLength) {

				free(payloadSizes);
				status->error = MMP_ERR_INVALID_PAYLOAD_COUNT;
				return;
			}
			break;
		}

		payloadSize++;
	}

	// TODO: Is this logic correct??? Am I checking the
	// right counter variable here?
	if (packet->payloadLen < payloadLength) {

		free(payloadSizes);
		status->error = MMP_ERR_INVALID_PAYLOAD_COUNT;
		return;
	}

	packet->payloads = (uint8_t **)malloc(sizeof(uint8_t *)*packet->payloadLen);
	if (packet->payloads == NULL) {

		status->error = MMP_ERR_MEMORY;
		free(payloadSizes);
		return;
	}

	payloadSize = PKT_IDX_PAYLOAD_START;

	for (int i = 0; i < packet->payloadLen; i++) {

		packet->payloads[i] = (uint8_t *)malloc(sizeof(uint8_t)*payloadSizes[i]);

		for (int j = 0; j < payloadSizes[i]; j++) {

			packet->payloads[i][j] = packet->bytes[start + j + payloadSize];
		}
		payloadSize += payloadSizes[i];
	}

	status->size = ++payloadSize;
	free(payloadSizes);
}
