/*	libmmp.h - M.A.R-T>Y. C lib protocol implementation header file
	Copyright (C) 2023-2024 hoff.industries

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <https://www.gnu.org/licenses/>. */

#include <stdint.h>
#include <stdbool.h>

const extern int MMP_LIB_VERSION[];
const extern int MMP_IMPLEMENTS[];

typedef uint8_t mmp_type;
typedef uint16_t mmp_arg;

#define LIBMMP_AUTHOR					"hoff.industries"

#define SONAR_RETRIES					3
#define SONAR_SLEEP_INTERVAL			5

#define MIN_SIZE						4
#define MAX_SIZE						4096
#define MAX_CLIENTS						20

#define DATA_EMPTY						0x00
#define ARGS_EMPTY						0x0000
#define TERM_PAYLOAD					0xFE
#define TERM_PACKET						0xFF

#define PKT_IDX_PAYLOAD_START			3

#define MMP_TYPE_UNSET					0x00
#define MMP_TYPE_INIT					0x01
#define MMP_TYPE_INVALID				0x02
#define MMP_TYPE_CONTERM				0x03
#define MMP_TYPE_SONAR					0x04
#define MMP_TYPE_SEND					0x05
#define MMP_TYPE_APPLICATION			0x06
#define MMP_TYPE_QUERY					0x07

#define MMP_ARG_UNSET					0x0000

#define MMP_ARG_INIT_INIT				0x0001 // client -> server
#define MMP_ARG_INIT_ACCEPT				0x0001 // server -> client

#define MMP_ARG_INVALID_TYPE			0x0001
#define MMP_ARG_INVALID_LENGTH			0x0002
#define MMP_ARG_INVALID_ARGS			0x0003
#define MMP_ARG_INVALID_PAYLOAD			0x0004
#define MMP_ARG_INVALID_PAYLOAD_COUNT	0x0005
#define MMP_ARG_INVALID_AUTH			0x0006 // server -> client
#define MMP_ARG_INVALID_KEY_EXISTS		0x0007 // server -> client
#define MMP_ARG_INVALID_KEY_RESERVED	0x0008
#define MMP_ARG_INVALID_MESSAGE_ID		0x0009 // server -> client

#define MMP_ARG_CONTERM_CLEAN			0x0001
#define MMP_ARG_CONTERM_SPAM			0x0002
#define MMP_ARG_CONTERM_AUTH			0x0003

#define MMP_SEND_DIRECT					0x0001
#define MMP_SEND_BROADCAST				0x0002

#define MMP_APPLICATION_SUCCESS			0x0001
#define MMP_APPLICATION_DELIVERY		0x0002
#define MMP_APPLICATION_BUSY			0x0003
#define MMP_APPLICATION_RANGE			0x0004
#define MMP_APPLICATION_ENDPOINT		0x0005
#define MMP_APPLICATION_VALUE			0x0006
#define MMP_APPLICATION_TOO_FEW			0x0007
#define MMP_APPLICATION_TOO_MANY		0x0008
#define MMP_APPLICATION_AUTH			0x0009
#define MMP_APPLICATION_EXECUTE			0x000A

#define MMP_QUERY_CONN_NUMBER			0x0001
#define MMP_QUERY_CLIENT_INFO			0x0002
#define MMP_QUERY_NOT_FOUND				0x0003 // server -> client

#define MMP_PLEN_UNSET					0
#define MMP_PLEN_INIT					4
#define MMP_PLEN_INVALID				1
#define MMP_PLEN_CONTERM				0
#define MMP_PLEN_SONAR					0
#define MMP_PLEN_SEND					5
#define MMP_PLEN_APPLICATION			1
#define MMP_PLEN_QUERY					3

enum PayloadField {
	OPT,
	OPT_R,
	REQ,
	EMPTY,
};

enum MMPError {
	MMP_ERR_NONE,
	MMP_ERR_NETWORK,
	MMP_ERR_MEMORY,
	MMP_ERR_FORMAT,

	MMP_ERR_INIT_FAIL,
	MMP_ERR_DELIVERY_FAIL,
	MMP_ERR_QUERY_FAIL,

	MMP_ERR_QUEUE_PUSH,
	MMP_ERR_QUEUE_PULL,

	MMP_ERR_UNSET_MESSAGE_TYPE,
	MMP_ERR_UNSET_ARGUMENT,
	MMP_ERR_UNSET_PAYLOADS,
	MMP_ERR_UNSET_PACKET,

	MMP_ERR_INVALID_TYPE,
	MMP_ERR_INVALID_ARGUMENT,
	MMP_ERR_INVALID_PACKET_LENGTH,
	MMP_ERR_INVALID_PAYLOAD_COUNT,
	MMP_ERR_INVALID_PAYLOAD_VALUE,
	MMP_ERR_INVALID_MAP_KEY,

	MMP_ERR_MESSAGE_ID_UNAVAILABLE,
	MMP_ERR_WAIT_INTERRUPTION,
	MMP_ERR_AUTH_FAIL,
};

enum ClientState {
	PREINIT,
	ACTIVE,
	INACTIVE
};

struct MMPacket {
	mmp_type type;
	mmp_arg argument;
	uint8_t **payloads;
	int payloadLen;

	uint8_t bytes[MAX_SIZE];
};

struct MMStatus {
	int size;
	enum MMPError error;
};

void print_mmpacket(struct MMPacket *packet);

void mmpacket_factory(struct MMPacket *packet, bool zero);
void mmpacket_reset(struct MMPacket *packet, bool zero);
void mmstatus_factory(struct MMStatus *status);

int mmpayloadlen(uint8_t *payload);
void str_2_mmpayload(char *string, uint8_t *payload);
void mmpayload_2_str(uint8_t *payload, char *string);
struct MMPacket *mmpacket_2_heap(struct MMPacket *packet, bool doSerial, int start);

void mmp_serialize(struct MMPacket *packet, struct MMStatus *status);
void mmp_deserialize(struct MMPacket *packet, struct MMStatus *status, int start);
